#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""Generic Tango Pool Device base classes"""

__all__ = ["PoolDevice", "PoolDeviceClass",
           "PoolElementDevice", "PoolElementDeviceClass",
           "PoolExpChannelDevice", "PoolExpChannelDeviceClass",
           "PoolGroupDevice", "PoolGroupDeviceClass"]

__docformat__ = 'restructuredtext'

import time

from PyTango import Util, DevVoid, DevLong64, DevBoolean, DevString,\
    DevDouble, DevEncoded, DevVarStringArray, DispLevel, DevState, SCALAR, \
    SPECTRUM, IMAGE, READ_WRITE, READ, AttrData, CmdArgType, DevFailed,\
    seqStr_2_obj, Except, ErrSeverity

from taurus.core.util.containers import CaselessDict
from taurus.core.util.codecs import CodecFactory

from sardana import InvalidId, InvalidAxis, ElementType
from sardana import sardanacustomsettings
from sardana.pool.poolmetacontroller import DataInfo
from sardana.tango.core.SardanaDevice import SardanaDevice, SardanaDeviceClass
from sardana.tango.core.util import GenericScalarAttr, GenericSpectrumAttr, \
    GenericImageAttr, to_tango_attr_info


class PoolDevice(SardanaDevice):
    """Base Tango Pool device class"""

    #: list of extreme error states
    ExtremeErrorStates = DevState.FAULT, DevState.UNKNOWN

    #: list of busy states
    BusyStates = DevState.MOVING, DevState.RUNNING

    #: Maximum number of retries in a busy state
    BusyRetries = 3

    def __init__(self, dclass, name):
        """Constructor"""
        SardanaDevice.__init__(self, dclass, name)

    def init(self, name):
        """initialize the device once in the object lifetime. Override when
        necessary but **always** call the method from your super class

        :param :obj:`str` name: device name"""
        SardanaDevice.init(self, name)
        util = Util.instance()
        self._pool_device = util.get_device_list_by_class("Pool")[0]
        self._element = None

    @property
    def pool_device(self):
        """The tango pool device"""
        return self._pool_device

    @property
    def pool(self):
        """The sardana pool object"""
        return self.pool_device.pool

    def get_element(self):
        """Returns the underlying pool element object

        :return: the underlying pool element object
        :rtype: :class:`~sardana.pool.poolelement.PoolElement`"""
        return self._element

    def set_element(self, element):
        """Associates this device with the sardana element

        :param element: the sardana element
        :type element: :class:`~sardana.pool.poolelement.PoolElement`"""
        self._element = element

    element = property(get_element, set_element,
                       doc="The underlying sardana element")

    def init_device(self):
        """Initialize the device. Called during startup after :meth:`init` and
        every time the tango ``Init`` command is executed.
        Override when necessary but **always** call the method from your super
        class"""
        SardanaDevice.init_device(self)

    def delete_device(self):
        """Clean the device. Called during shutdown and every time the tango
        ``Init`` command is executed.
        Override when necessary but **always** call the method from your super
        class"""
        SardanaDevice.delete_device(self)

    def Abort(self):
        """The tango abort command. Aborts the active operation"""
        self.element.abort()
        try:
            self.element.get_state(cache=False, propagate=2)
        except:
            self.warning("Abort: failed to read state")

    def is_Abort_allowed(self):
        """Returns True if it is allowed to execute the tango abort command

        :return: True if it is allowed to execute the tango abort command or
                 False otherwise
        :rtype: bool"""
        return self.get_state() != DevState.UNKNOWN

    def Stop(self):
        """The tango stop command. Stops the active operation"""
        self.element.stop()
        try:
            self.element.get_state(cache=False, propagate=2)
        except:
            self.info("Stop: failed to read state")

    def is_Stop_allowed(self):
        """Returns True if it is allowed to execute the tango stop command

        :return: True if it is allowed to execute the tango stop command or
                 False otherwise
        :rtype: bool"""
        return self.get_state() != DevState.UNKNOWN

    def _is_allowed(self, req_type):
        """Generic is_allowed"""
#        state = self.get_state()
#        if state in self.ExtremeErrorStates:
#            return False
#        if req_type == AttReqType.WRITE_REQ:
#            if state in self.BusyStates:
#                return False
        return True

    def get_dynamic_attributes(self):
        """Returns the standard dynamic and fully dynamic attributes for this
        device. The return is a tuple of two dictionaries:

        - standard attributes: caseless dictionary with key being the attribute
          name and value is a tuple of attribute name(str), tango information,
          attribute information
        - dynamic attributes: caseless dictionary with key being the attribute
          name and value is a tuple of attribute name(str), tango information,
          attribute information

        **tango information**
            seq< :class:`~PyTango.CmdArgType`, :class:`~PyTango.AttrDataFormat`, :class:`~PyTango.AttrWriteType` >

        **attribute information**
            attribute information as returned by the sardana controller

        :return: the standard dynamic and fully dynamic attributes
        :rtype: seq< :class:`~taurus.core.util.CaselessDict`, :class:`~taurus.core.util.CaselessDict`\>
        """
        return CaselessDict(), CaselessDict()

    def initialize_dynamic_attributes(self):
        """Initializes this device dynamic attributes"""
        self._attributes = attrs = CaselessDict()

        attr_data = self.get_dynamic_attributes()

        std_attrs, dyn_attrs = attr_data
        self.remove_unwanted_dynamic_attributes(std_attrs, dyn_attrs)

        if std_attrs is not None:
            read = self.__class__._read_DynamicAttribute
            write = self.__class__._write_DynamicAttribute
            is_allowed = self.__class__._is_DynamicAttribute_allowed
            for attr_name, data_info in list(std_attrs.items()):
                attr_name, data_info, attr_info = data_info
                attr = self.add_standard_attribute(attr_name, data_info,
                                                   attr_info, read,
                                                   write, is_allowed)
                attrs[attr.get_name()] = None

        if dyn_attrs is not None:
            read = self.__class__._read_DynamicAttribute
            write = self.__class__._write_DynamicAttribute
            is_allowed = self.__class__._is_DynamicAttribute_allowed
            for attr_name, data_info in list(dyn_attrs.items()):
                attr_name, data_info, attr_info = data_info
                attr = self.add_dynamic_attribute(attr_name, data_info,
                                                  attr_info, read,
                                                  write, is_allowed)
                attrs[attr.get_name()] = None
        return attrs

    def remove_unwanted_dynamic_attributes(self, new_std_attrs, new_dyn_attrs):
        """Removes unwanted dynamic attributes from previous device creation"""

        dev_class = self.get_device_class()
        multi_attr = self.get_device_attr()
        multi_class_attr = dev_class.get_class_attr()
        static_attr_names = \
            list(map(str.lower, list(dev_class.attr_list.keys())))
        static_attr_names.extend(('state', 'status'))

        new_attrs = CaselessDict(new_std_attrs)
        new_attrs.update(new_dyn_attrs)

        device_attr_names = []
        for i in range(multi_attr.get_attr_nb()):
            device_attr_names.append(multi_attr.get_attr_by_ind(i).get_name())

        for attr_name in device_attr_names:
            attr_name_lower = attr_name.lower()
            if attr_name_lower in static_attr_names:
                continue
            try:
                self.remove_attribute(attr_name)
            except:
                self.warning("Error removing dynamic attribute %s",
                             attr_name_lower)
                self.debug("Details:", exc_info=1)

        klass_attr_names = []
        klass_attrs = multi_class_attr.get_attr_list()
        for ind in range(len(klass_attrs)):
            klass_attr_names.append(klass_attrs[ind].get_name())

        for attr_name in klass_attr_names:
            attr_name_lower = attr_name.lower()
            if attr_name_lower in static_attr_names:
                continue
            # if new dynamic attribute is in class attribute then delete it
            # from class attribute to be later on added again (eventually
            # with diffent data type or data format)
            if attr_name_lower in new_attrs:
                try:
                    attr = multi_class_attr.get_attr(attr_name)

                    old_type = CmdArgType(attr.get_type())
                    old_format = attr.get_format()
                    old_access = attr.get_writable()

                    new_attr = new_attrs[attr_name]
                    new_type, new_format, new_access = new_attr[1][0][:3]
                    differ = new_type != old_type or \
                        new_format != old_format or \
                        new_access != old_access
                    if differ:
                        self.info("Replacing dynamic attribute %s", attr_name)
                        self.debug("old type: %s, new type: %s",
                                   old_type, new_type)
                        self.debug("old format: %s, new format: %s",
                                   old_format, new_format)
                        self.debug("old access: %s, new access: %s",
                                   old_access, new_access)
                        multi_class_attr.remove_attr(attr.get_name(),
                                                     attr.get_cl_name())
                except:
                    self.warning("Error removing dynamic attribute %s from "
                                 " device class", attr_name)
                    self.debug("Details:", exc_info=1)

    def add_dynamic_attribute(self, attr_name, data_info, attr_info, read,
                              write, is_allowed):
        """Adds a single dynamic attribute

        :param :obj:`str` attr_name: the attribute name
        :param data_info: tango attribute information
        :type data_info: seq< :class:`~PyTango.CmdArgType`, :class:`~PyTango.AttrDataFormat`, :class:`~PyTango.AttrWriteType` >
        :param attr_info: attribute information
        :param read: read method for the attribute
        :param write: write method for the attribute
        :param is_allowed: is allowed method"""
        tg_type, tg_format, tg_access = data_info[0]

        if tg_access == READ:
            write = None
        if tg_format == SCALAR:
            attr = GenericScalarAttr(attr_name, tg_type, tg_access)
        if tg_format == SPECTRUM:
            dim_x = attr_info.maxdimsize[0]
            attr = GenericSpectrumAttr(attr_name, tg_type, tg_access,
                                       dim_x=dim_x)
        elif tg_format == IMAGE:
            dim_x, dim_y = attr_info.maxdimsize
            attr = GenericImageAttr(attr_name, tg_type, tg_access,
                                    dim_x=dim_x, dim_y=dim_y)

        if tg_access == READ_WRITE and tg_format == SCALAR:
            memorized = attr_info.memorized.lower()
            if memorized == 'true':
                attr.set_memorized()
                attr.set_memorized_init(True)
            elif memorized == 'true_without_hard_applied':
                attr.set_memorized()
                attr.set_memorized_init(False)
        attr.set_disp_level(DispLevel.EXPERT)
        return self.add_attribute(attr, read, write, is_allowed)

    def add_standard_attribute(self, attr_name, data_info, attr_info, read,
                               write, is_allowed):
        """Adds a single standard dynamic attribute

        :param :obj:`str` attr_name: the attribute name
        :param data_info: tango attribute information
        :type data_info: seq< :class:`~PyTango.CmdArgType`, :class:`~PyTango.AttrDataFormat`, :class:`~PyTango.AttrWriteType` >
        :param attr_info: attribute information
        :param read: read method for the attribute
        :param write: write method for the attribute
        :param is_allowed: is allowed method"""
        dev_class = self.get_device_class()
        attr_data = AttrData(attr_name, dev_class.get_name(), data_info)
        attr = self.add_attribute(attr_data, read, write, is_allowed)
        return attr

    def read_DynamicAttribute(self, attr):
        """Generic read dynamic attribute.
        Default implementation raises :exc:`NotImplementedError`

        :param attr: attribute to be read
        :type attr: :class:`~PyTango.Attribute`

        :raises: :exc:`NotImplementedError`"""
        raise NotImplementedError

    def write_DynamicAttribute(self, attr):
        """Generic write dynamic attribute.
        Default implementation raises :exc:`NotImplementedError`

        :param attr: attribute to be written
        :type attr: :class:`~PyTango.Attribute`

        :raises: :exc:`NotImplementedError`"""
        raise NotImplementedError

    def is_DynamicAttribute_allowed(self, req_type):
        """Generic is dynamic attribute allowed.
        Default implementation calls :meth:`_is_allowed`

        :param req_type: request type
        :type attr: :class:`~PyTango.AttrRequestType`"""
        return self._is_allowed(req_type)

    def _read_DynamicAttribute(self, attr):
        """Generic internal read dynamic attribute.
        Checks if this object has a 'read_'+<attr_name> method and calls it.
        If not calls :meth:`read_DynamicAttribute`.

        :param attr: attribute to be read
        :type attr: :class:`~PyTango.Attribute`"""
        name = attr.get_name()

        read_name = "read_" + name
        if hasattr(self, read_name):
            read = getattr(self, read_name)
            return read(attr)

        return self.read_DynamicAttribute(attr)

    def _write_DynamicAttribute(self, attr):
        """Generic internal write dynamic attribute.
        Checks if this object has a 'write_'+<attr_name> method and calls it.
        If not calls :meth:`write_DynamicAttribute`.

        :param attr: attribute to be written
        :type attr: :class:`~PyTango.Attribute`"""
        name = attr.get_name()
        write_name = "write_" + name
        if hasattr(self, write_name):
            write = getattr(self, write_name)
            return write(attr)
        return self.write_DynamicAttribute(attr)

    def _is_DynamicAttribute_allowed(self, req_type):
        """Generic is dynamic attribute allowed.
        Default implementation calls :meth:`is_DynamicAttribute_allowed`

        :param req_type: request type
        :type attr: :class:`~PyTango.AttrRequestType`"""
        return self.is_DynamicAttribute_allowed(req_type)

    def dev_state(self):
        """Calculates and returns the device state. Called by Tango on a read
        state request.

        :return: the device state
        :rtype: :class:`~PyTango.DevState`"""
        element = self.element
        try:
            use_cache = element.is_in_operation() and not self.Force_HW_Read
            ctrl_state = element.get_state(cache=use_cache, propagate=0)
            state = self.calculate_tango_state(ctrl_state)
            return state
        except:
            self.error("Exception trying to return state")
            self.debug("Details:", exc_info=1)
            return DevState.FAULT

    def dev_status(self):
        """
        Calculates and returns the device status. Called by Tango on a read
        status request.

        :return: the device status
        :rtype: :obj:`str`
        """
        element = self.element
        try:
            use_cache = element.is_in_operation() and not self.Force_HW_Read
            ctrl_status = self.element.get_status(cache=use_cache, propagate=0)
            status = self.calculate_tango_status(ctrl_status)
            return status
        except Exception as e:
            msg = "Exception trying to return status: %s" % str(e)
            self.error(msg)
            self.debug("Details:", exc_info=1)
            return msg

    def wait_for_operation(self):
        """Waits for an operation to finish. It uses the maxumum number of
        retries. Sleeps 0.01s between retries.

        :raises: :exc:`Exception` in case of a timeout"""
        element, n = self.element, self.BusyRetries
        while element.is_in_operation():
            if n == 0:
                raise Exception("Wait for operation timedout")
            time.sleep(0.01)
            self.warning("waited for operation")
            n = n - 1

    def Restore(self):
        """Restore tango command. Restores the attributes to their former glory.
        This applies to memorized writable attributes which have a set point
        stored in the database"""
        restore_attributes, db_values = self.get_restore_data()
        multi_attribute = self.get_device_attr()

        for attr_name in restore_attributes:
            props = db_values[attr_name]
            if props is None or not "__value" in props:
                continue
            attribute = multi_attribute.get_w_attr_by_name(attr_name)
            write_meth_name = "write_" + attr_name
            write_meth = getattr(self, write_meth_name, None)
            if write_meth is None:
                self.warning("Could not recover %s: %s does not exist",
                             attr_name, write_meth_name)
                continue
            self.restore_attribute(attribute, write_meth, props['__value'])

    def get_restore_data(self):
        restore_attributes = self.get_attributes_to_restore()
        db = Util.instance().get_database()
        db_values = db.get_device_attribute_property(self.get_name(),
                                                     restore_attributes)
        return restore_attributes, db_values

    def get_attributes_to_restore(self):
        std_attrs, dyn_attrs = self.get_dynamic_attributes()
        multi_attribute = self.get_device_attr()

        restore = []
        for attr_name in std_attrs:
            try:
                attribute = multi_attribute.get_w_attr_by_name(attr_name)
            except DevFailed:
                continue
            restore.append(attribute.get_name())
        for attr_name in dyn_attrs:
            try:
                attribute = multi_attribute.get_w_attr_by_name(attr_name)
            except DevFailed:
                continue
            restore.append(attribute.get_name())
        return restore

    def _get_attribute_value_from_db_value(self, attribute, db_value):
        value = seqStr_2_obj(db_value, attribute.get_data_type(),
                             attribute.get_data_format())
        return value

    def restore_attribute(self, attribute, write_meth, db_value):
        value = self._get_attribute_value_from_db_value(attribute, db_value)
        attr_name = attribute.get_name()
        try:
            attribute.set_write_value(value)
            self.info("Restoring %s", attr_name)
            write_meth(attribute)
        except:
            self.warning("Could not recover %s: Error in write", attr_name)
            self.debug("Details:", exc_info=1)


class PoolDeviceClass(SardanaDeviceClass):
    """Base Tango Pool Device Class class"""

    #:
    #: Sardana device class properties definition
    #:
    #: .. seealso:: :ref:`server`
    #:
    class_property_list = SardanaDeviceClass.class_property_list

    #:
    #: Sardana device properties definition
    #:
    #: .. seealso:: :ref:`server`
    #:
    device_property_list = {
        'Id': [DevLong64, "Internal ID", InvalidId],
        'Force_HW_Read': [DevBoolean, "Force a hardware read of value even "
                          "when in operation (motion/acquisition",
                          False],
    }
    device_property_list.update(SardanaDeviceClass.device_property_list)

    #:
    #: Sardana device command definition
    #:
    #: .. seealso:: :ref:`server`
    #:
    cmd_list = {
        'Stop': [[DevVoid, ""], [DevVoid, ""]],
        'Abort': [[DevVoid, ""], [DevVoid, ""]],
        'Restore': [[DevVoid, ""], [DevVoid, ""]],
    }
    cmd_list.update(SardanaDeviceClass.cmd_list)

    #:
    #: Sardana device attribute definition
    #:
    #: .. seealso:: :ref:`server`
    #:
    attr_list = {
    }
    attr_list.update(SardanaDeviceClass.attr_list)

    standard_attr_list = {}

    def _get_class_properties(self):
        ret = SardanaDeviceClass._get_class_properties(self)
        ret['Description'] = "Generic Pool device class"
        ret['InheritedFrom'].insert(0, 'SardanaDevice')
        return ret


class PoolElementDevice(PoolDevice):
    """Base Tango Pool Element Device class"""

    def init_device(self):
        """Initialize the device. Called during startup after :meth:`init` and
        every time the tango ``Init`` command is executed.
        Override when necessary but **always** call the method from your super
        class"""
        PoolDevice.init_device(self)

        self.instrument = None
        self.ctrl = None
        try:
            instrument_id = int(self.Instrument_id)
            if instrument_id != InvalidId:
                instrument = self.pool.get_element_by_id(instrument_id)
                self.instrument = instrument
        except ValueError:
            pass
        try:
            ctrl_id = int(self.Ctrl_id)
            if ctrl_id != InvalidId:
                ctrl = self.pool.get_element_by_id(ctrl_id)
                self.ctrl = ctrl
        except ValueError:
            pass

    def delete_device(self):
        element = self.element
        if not element.deleted:
            pool_ctrl = self.ctrl
            ctrl = pool_ctrl.ctrl
            if ctrl is not None:
                axis = element.axis
                try:
                    ctrl.DeleteDevice(axis)
                except Exception:
                    name = element.name
                    self.error("Unable to delete %s(%s)", name, axis,
                               exc_info=1)
        PoolDevice.delete_device(self)

    def read_Instrument(self, attr):
        """Read the value of the ``Instrument`` tango attribute.
        Returns the instrument full name or empty string if this element doesn't
        belong to any instrument

        :param attr: tango instrument attribute
        :type attr: :class:`~PyTango.Attribute`"""
        instrument = self.element.instrument
        if instrument is None:
            attr.set_value('')
        else:
            attr.set_value(instrument.full_name)

    def write_Instrument(self, attr):
        """Write the value of the ``Instrument`` tango attribute.
        Sets a new instrument full name or empty string if this element doesn't
        belong to any instrument.
        The instrument **must** have been previously created.

        :param attr: tango instrument attribute
        :type attr: :class:`~PyTango.Attribute`"""
        name = attr.get_write_value()
        instrument = None
        if name:
            instrument = self.pool.get_element(full_name=name)
            if instrument.get_type() != ElementType.Instrument:
                raise Exception("%s is not an instrument" % name)
        self.element.instrument = instrument
        db = Util.instance().get_database()
        db.put_device_property(
            self.get_name(), {"Instrument_id": instrument.id})

    def get_dynamic_attributes(self):
        """Override of :class:`PoolDevice.get_dynamic_attributes`.
        Returns the standard dynamic and fully dynamic attributes for this
        device. The return is a tuple of two dictionaries:

        - standard attributes: caseless dictionary with key being the attribute
          name and value is a tuple of attribute name(str), tango information,
          attribute information
        - dynamic attributes: caseless dictionary with key being the attribute
          name and value is a tuple of attribute name(str), tango information,
          attribute information

        **tango information**
            seq< :class:`~PyTango.CmdArgType`, :class:`~PyTango.AttrDataFormat`, :class:`~PyTango.AttrWriteType` >

        **attribute information**
            attribute information as returned by the sardana controller

        :return: the standard dynamic and fully dynamic attributes
        :rtype: seq< :class:`~taurus.core.util.CaselessDict`, :class:`~taurus.core.util.CaselessDict`\>
        """

        if hasattr(self, "_dynamic_attributes_cache"):
            return self._standard_attributes_cache, \
                   self._dynamic_attributes_cache
        std_attrs, dyn_attrs = self._get_dynamic_attributes()
        self._standard_attributes_cache = std_attrs
        self._dynamic_attributes_cache = dyn_attrs
        return std_attrs, dyn_attrs

    def _get_dynamic_attributes(self):
        ctrl = self.ctrl
        if ctrl is None:
            self.warning("no controller: dynamic attributes NOT created")
            return PoolDevice.get_dynamic_attributes(self)
        if not ctrl.is_online():
            self.warning("controller offline: dynamic attributes NOT created")
            return PoolDevice.get_dynamic_attributes(self)

        dyn_attrs = CaselessDict()
        std_attrs = CaselessDict()
        dev_class = self.get_device_class()
        axis_attrs = ctrl.get_axis_attributes(self.element.axis)

        std_attrs_lower = [attr.lower()
                           for attr in dev_class.standard_attr_list]
        for attr_name, attr_info in list(axis_attrs.items()):
            attr_name_lower = attr_name.lower()
            if attr_name_lower in std_attrs_lower:
                data_info = DataInfo.toDataInfo(attr_name, attr_info)
                tg_info = dev_class.standard_attr_list[attr_name]
                std_attrs[attr_name] = attr_name, tg_info, data_info
            else:
                data_info = DataInfo.toDataInfo(attr_name, attr_info)
                name, tg_info = to_tango_attr_info(attr_name, data_info)
                dyn_attrs[attr_name] = name, tg_info, data_info
        return std_attrs, dyn_attrs

    def read_DynamicAttribute(self, attr):
        """Read a generic dynamic attribute. Calls the controller of this
        element to get the dynamic attribute value

        :param attr: tango attribute
        :type attr: :class:`~PyTango.Attribute`"""
        name = attr.get_name()
        ctrl = self.ctrl
        if ctrl is None:
            raise Exception("Cannot read %s. Controller not build!" % name)
        v = ctrl.get_axis_attr(self.element.axis, name)
        if v is None:
            raise TypeError(
                "Cannot read %s. Controller returns None" % (name,))
        attr.set_value(v)

    def write_DynamicAttribute(self, attr):
        """Write a generic dynamic attribute. Calls the controller of this
        element to get the dynamic attribute value

        :param attr: tango attribute
        :type attr: :class:`~PyTango.Attribute`"""
        name = attr.get_name()
        value = attr.get_write_value()
        self.debug("writing dynamic attribute %s with value %s", name, value)
        ctrl = self.ctrl
        if ctrl is None:
            raise Exception("Cannot write %s. Controller not build!" % name)
        ctrl.set_axis_attr(self.element.axis, name, value)

    def read_SimulationMode(self, attr):
        """Read the current simulation mode.

        :param attr: tango attribute
        :type attr: :class:`~PyTango.Attribute`"""
        attr.set_value(self.element.simulation_mode)

    def write_SimulationMode(self, attr):
        """Sets the simulation mode.

        :param attr: tango attribute
        :type attr: :class:`~PyTango.Attribute`"""
        self.element.simulation_mode = attr.get_write_value()


class PoolElementDeviceClass(PoolDeviceClass):
    """Base Tango Pool Element Device Class class"""

    #:
    #: Sardana device properties definition
    #:
    #: .. seealso:: :ref:`server`
    #:
    device_property_list = {
        "Axis": [DevLong64, "Axis in the controller", [InvalidAxis]],
        "Ctrl_id": [DevLong64, "Controller ID", [InvalidId]],
        "Instrument_id": [DevLong64, "Controller ID", [InvalidId]],
    }
    device_property_list.update(PoolDeviceClass.device_property_list)

    #:
    #: Sardana device attribute definition
    #:
    #: .. seealso:: :ref:`server`
    #:
    attr_list = {
        'Instrument':    [[DevString, SCALAR, READ_WRITE],
                          {'label': "Instrument",
                           'Display level': DispLevel.EXPERT}],
        'SimulationMode': [[DevBoolean, SCALAR, READ_WRITE],
                           {'label': "Simulation mode"}],
    }
    attr_list.update(PoolDeviceClass.attr_list)

    cmd_list = {
    }
    cmd_list.update(PoolDeviceClass.cmd_list)

    def get_standard_attr_info(self, attr):
        """Returns information about the standard attribute

        :param :obj:`str` attr: attribute name
        :return: a sequence of tango data_type, data format"""
        return self.standard_attr_list[attr]

    def _get_class_properties(self):
        ret = PoolDeviceClass._get_class_properties(self)
        ret['Description'] = "Generic Pool element device class"
        ret['InheritedFrom'].insert(0, 'PoolDevice')
        return ret


class PoolGroupDevice(PoolDevice):
    """Base Tango Pool Group Device class"""

    def read_ElementList(self, attr):
        """Read the element list.

        :param attr: tango attribute
        :type attr: :class:`~PyTango.Attribute`"""
        attr.set_value(self.get_element_names())

    def get_element_names(self):
        """Returns the list of element names.

        :return: a list of attribute names"""
        elements = self.element.get_user_elements()
        return [element.name for element in elements]

    def elements_changed(self, evt_src, evt_type, evt_value):
        """Callback for when the elements of this group changed"""
        self.push_change_event("ElementList", self.get_element_names())


class PoolGroupDeviceClass(PoolDeviceClass):
    """Base Tango Pool Group Device Class class"""

    #:
    #: Sardana device properties definition
    #:
    #: .. seealso:: :ref:`server`
    #:
    device_property_list = {
        "Elements":    [DevVarStringArray, "elements in the group", []],
    }
    device_property_list.update(PoolDeviceClass.device_property_list)

    #:
    #: Sardana device command definition
    #:
    #: .. seealso:: :ref:`server`
    #:
    cmd_list = {
    }
    cmd_list.update(PoolDeviceClass.cmd_list)

    #:
    #: Sardana device attribute definition
    #:
    #: .. seealso:: :ref:`server`
    #:
    attr_list = {
        'ElementList': [[DevString, SPECTRUM, READ, 4096]],
    }
    attr_list.update(PoolDeviceClass.attr_list)

    def _get_class_properties(self):
        ret = PoolDeviceClass._get_class_properties(self)
        ret['Description'] = "Generic Pool group device class"
        ret['InheritedFrom'].insert(0, 'PoolDevice')
        return ret


class PoolExpChannelDevice(PoolElementDevice):

    def __init__(self, dclass, name):
        """Constructor"""
        PoolElementDevice.__init__(self, dclass, name)
        codec_name = getattr(sardanacustomsettings, "VALUE_BUFFER_CODEC")
        self._value_buffer_codec = CodecFactory().getCodec(codec_name)
        codec_name = getattr(sardanacustomsettings, "VALUE_REF_BUFFER_CODEC")
        self._value_ref_buffer_codec = CodecFactory().getCodec(codec_name)

    def _encode_value_chunk(self, value_chunk):
        """Prepare value chunk to be passed via communication channel.

        :param value_chunk: value chunk
        :type value_chunk: seq<SardanaValue>

        :return: json string representing value chunk
        :rtype: str"""
        index = []
        value = []
        for idx, sdn_value in value_chunk.items():
            index.append(idx)
            value.append(sdn_value.value)
        data = dict(index=index, value=value)
        encoded_data = self._value_buffer_codec.encode(('', data))
        return encoded_data

    def _encode_value_ref_chunk(self, value_ref_chunk):
        """Prepare value ref chunk to be passed via communication channel.

        :param value_ref_chunk: value ref chunk
        :type value_ref_ chunk: seq<SardanaValue>

        :return: json string representing value chunk
        :rtype: str
        """
        index = []
        value_ref = []
        for idx, sdn_value in value_ref_chunk.items():
            index.append(idx)
            value_ref.append(sdn_value.value)
        data = dict(index=index, value_ref=value_ref)
        encoded_data = self._value_ref_buffer_codec.encode(('', data))
        return encoded_data

    def initialize_dynamic_attributes(self):
        attrs = PoolElementDevice.initialize_dynamic_attributes(self)

        non_detect_evts = "integrationtime",

        for attr_name in non_detect_evts:
            if attr_name in attrs:
                self.set_change_event(attr_name, True, False)
        return attrs

    def read_ValueBuffer(self, _):
        desc = "ValueBuffer attribute is not foreseen for reading. It is " \
               "used only as the communication channel for the continuous " \
               "acquisitions."
        Except.throw_exception("UnsupportedFeature",
                               desc,
                               "PoolExpChannelDevice.read_ValueBuffer",
                               ErrSeverity.WARN)

    def read_ValueRefBuffer(self, _):
        desc = ("ValueRefBuffer attribute is not foreseen for reading. "
                "It is used only as the communication channel for the "
                "continuous acquisitions.")
        Except.throw_exception("UnsupportedFeature",
                               desc,
                               "PoolExpChannelDevice.read_ValueRefBuffer",
                               ErrSeverity.WARN)

    def read_IntegrationTime(self, attr):
        """Reads the integration time.

        :param attr: tango attribute
        :type attr: :class:`~PyTango.Attribute`"""
        attr.set_value(self.element.integration_time)

    def write_IntegrationTime(self, attr):
        """Sets the integration time.

        :param attr: tango attribute
        :type attr: :class:`~PyTango.Attribute`"""
        self.element.integration_time = attr.get_write_value()


class PoolExpChannelDeviceClass(PoolElementDeviceClass):

    #:
    #: Sardana device attribute definition
    #:
    #: .. seealso:: :ref:`server`
    #:
    attr_list = {
        'IntegrationTime': [[DevDouble, SCALAR, READ_WRITE]]
    }
    attr_list.update(PoolElementDeviceClass.attr_list)

    standard_attr_list = {
        'ValueBuffer': [[DevEncoded, SCALAR, READ]]
    }
    standard_attr_list.update(PoolElementDeviceClass.standard_attr_list)


class PoolTimerableDevice(PoolExpChannelDevice):

    def __init__(self, dclass, name):
        """Constructor"""
        PoolExpChannelDevice.__init__(self, dclass, name)

    def initialize_dynamic_attributes(self):
        attrs = PoolExpChannelDevice.initialize_dynamic_attributes(self)

        detect_evts = "timer",

        for attr_name in detect_evts:
            if attr_name in attrs:
                self.set_change_event(attr_name, True, True)
        return attrs

    def read_Timer(self, attr):
        """Reads the timer for this channel.

        :param attr: tango attribute
        :type attr: :class:`~PyTango.Attribute`"""
        timer = self.element.timer
        if timer is None:
            timer = 'None'
        attr.set_value(timer)

    def write_Timer(self, attr):
        """Sets the timer for this channel.

        :param attr: tango attribute
        :type attr: :class:`~PyTango.Attribute`"""
        timer = attr.get_write_value()
        if timer == 'None':
            timer = None
        self.element.timer = timer


class PoolTimerableDeviceClass(PoolExpChannelDeviceClass):

    #:
    #: Sardana device attribute definition
    #:
    #: .. seealso:: :ref:`server`
    #:

    #    Attribute definitions
    attr_list = {
        'Timer': [[DevString, SCALAR, READ_WRITE],
                  {'Memorized': "true", }]
    }
    attr_list.update(PoolExpChannelDeviceClass.attr_list)

    standard_attr_list = {}
    standard_attr_list.update(PoolExpChannelDeviceClass.standard_attr_list)



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""Tests Read Position from Sardana using PyTango"""
import PyTango
import unittest
from sardana.tango.pool.test import BasePoolTestCase
from sardana.tango.core.util import get_free_alias
import numbers


class ReadMotorPositionOutsideLim(BasePoolTestCase, unittest.TestCase):
    """TestCase class for testing that read position is possible when
    motor is out of SW limits. Verify that position has a numeric type.
    """

    def setUp(self):
        """Create dummy motor controller and dummy motor element
        """
        super(ReadMotorPositionOutsideLim, self).setUp()
        sar_type = 'Motor'
        lib = 'DummyMotorController'
        cls = 'DummyMotorController'
        self.ctrl_name = get_free_alias(PyTango.Database(), "readposctrl")
        self.pool.CreateController([sar_type, lib, cls, self.ctrl_name])
        self.elem_name = get_free_alias(PyTango.Database(), "mot_test")
        axis = 1
        self.pool.CreateElement([sar_type, self.ctrl_name, str(axis),
                                 self.elem_name])
        self.elem = PyTango.DeviceProxy(self.elem_name)
        self.elem.DefinePosition(0)

    def test_read_position_outside_sw_lim(self):
        """Test bug #238: reading position when motor is out of SW lims.
        Verify that position has a numeric type."""
        pc = self.elem.get_attribute_config("position")
        pc.min_value = "1"
        pc.max_value = "2"
        self.elem.set_attribute_config(pc)
        try:
            posread = self.elem.read_attribute('position').value
        except Exception as e_read:
            msg = ("Position cannot be read. Exception: %s" % e_read)
            self.fail(msg)
        msg = ("Position is not a number")
        self.assertIsInstance(posread, numbers.Number, msg)

    def tearDown(self):
        """Remove motor element and motor controller
        """
        self.pool.DeleteElement(self.elem_name)
        self.pool.DeleteElement(self.ctrl_name)
        super(ReadMotorPositionOutsideLim, self).tearDown()



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module is part of the Python Sardana libray. It defines the base class
for Sardana manager"""



__all__ = ["SardanaElementManager", "SardanaIDManager"]

__docformat__ = 'restructuredtext'

from taurus.core.util.codecs import CodecFactory

from sardana import InvalidId


class SardanaElementManager(object):
    """A class capable of manage elements"""

    SerializationProtocol = 'json'

    def get_serialization_protocol(self):
        return self.SerializationProtocol

    def set_serialization_protocol(self, protocol):
        self.SerializationProtocol = protocol

    serialization_protocol = property(get_serialization_protocol,
                                      set_serialization_protocol,
                                      doc="the serialization protocol")

    def serialize_element(self, element, *args, **kwargs):
        obj = element.serialize(*args, **kwargs)
        return self.serialize_object(obj)

    def serialize_object(self, obj, *args, **kwargs):
        return CodecFactory().encode(self.serialization_protocol, ('', obj),
                                     *args, **kwargs)[1]

    def str_element(self, element, *args, **kwargs):
        obj = element.serialize(*args, **kwargs)
        return self.str_object(obj)

    def str_object(self, obj, *args, **kwargs):
        # TODO: use the active codec instead of hardcoded json
        return CodecFactory().encode('json', ('', obj), *args, **kwargs)[1]


class SardanaIDManager(object):
    """A class capable of manage ids"""

    _last_id = InvalidId

    def get_new_id(self):
        """Returns a new ID. The ID becomes reserved at this moment.

        :return: a new ID
        :rtype: int"""
        self._last_id += 1
        return self._last_id

    def rollback_id(self):
        """Free previously reserved ID"""
        self._last_id -= 1

    def reserve_id(self, nid):
        """Marks the given ID as reserved

        :param id: the ID to be reserved
        :type id: int"""
        assert isinstance(nid, int)
        if nid > self._last_id:
            self._last_id = nid



import sys
import os
import imp
import PyTango


class Upgrade:

    def upgrade(self, db, serv, old, new):
        raise RuntimeError(
            "Invalid Upgrade class. Must define a upgrade method")

    @classmethod
    def fromTo(cls):
        raise RuntimeError(
            "Invalid Upgrade class. Must define a fromTo method")

    @classmethod
    def supports(cls, old, new):
        raise RuntimeError(
            "Invalid Upgrade class. Must define a supports method")

    @classmethod
    def supports_old(cls, old):
        raise RuntimeError(
            "Invalid Upgrade class. Must define a supports_old method")

    @classmethod
    def supports_new(cls, new):
        raise RuntimeError(
            "Invalid Upgrade class. Must define a supports_new method")


def print_list(l):
    for i in l:
        print i


def get_db(host=None, port=None):
    if host is None or port is None:
        return PyTango.Database()
    return PyTango.Database(host, port)


def get_server_list(type="Pool", host=None, port=None, db=None):
    type += '/'
    db = db or get_db(host, port)
    return [s[s.index('/') + 1:] for s in db.get_server_list() if s.startswith(type)]


def print_server_list():
    for s in get_server_list():
        print s


def get_pool_device_from_server(serv, host=None, port=None, db=None):
    db = db or get_db(host, port)
    return db.get_device_name("Pool/%s" % serv, "Pool")[0]


def get_pool_server_version(serv, host=None, port=None, db=None):
    db = db or get_db(host, port)
    pool_dev_name = get_pool_device_from_server(serv, db=db)
    if not len(db.get_device_property_list(pool_dev_name, "Version")):
        # if not in database it means it is version 0.1.0
        return "0.1.0"
    else:
        return db.get_device_property(pool_dev_name, ["Version"])["Version"][0]


def to_version_nb(v_str):
    m, n, r = v_str.split('.')
    return int(m) * 10000 + int(n) * 100 + int(r)


def strip_old_versions(v_list, thresold):
    thresold = to_version_nb(thresold)
    return [v for v in v_list if to_version_nb(v) > thresold]


def get_suitable_upgrade(old_vers, new_vers):
    upgrades = get_supported_upgrades()
    for u_kcls in upgrades:
        if u_kcls.supports(old_vers, new_vers):
            return u_kcls
    return None


def get_supported_upgrades():
    upgrade_classes = []
    curr_dir = os.path.abspath(os.path.dirname(__file__))
    for mod_name in os.listdir(curr_dir):
        if not mod_name.endswith(".py"):
            continue
        mod_name = mod_name[:mod_name.rindex(".py")]
        if mod_name == __name__:
            continue
        if mod_name.startswith("upgrade_"):
            continue
        m = imp.load_module(mod_name, *imp.find_module(mod_name))
        for s in m.__dict__.values():
            if isinstance(s, type(Upgrade)):
                if s == Upgrade:
                    continue
                if issubclass(s, Upgrade):
                    upgrade_classes.append(s)

    return upgrade_classes


def get_possible_upgrades(serv, host=None, port=None, db=None):
    curr_vers = get_pool_server_version(serv, host=host, port=port, db=db)
    upgrades = get_supported_upgrades()
    possible_upgrades = []
    for u_kcls in upgrades:
        if u_kcls.supports_old(curr_vers):
            possible_upgrades.append(u_kcls.fromTo()[1])
    return possible_upgrades



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

""" """

__docformat__ = 'restructuredtext'

import PyTango.constants


def __init_pytango_devfailed(mod):
    consts = (
        "API_AttrConfig", "API_AttrEventProp", "API_AttrIncorrectDataNumber",
        "API_AttrNoAlarm", "API_AttrNotAllowed", "API_AttrNotFound",
        "API_AttrNotWritable", "API_AttrOptProp", "API_AttrPropValueNotSet",
        "API_AttrValueNotSet", "API_AttrWrongDefined", "API_AttrWrongMemValue",
        "API_BadConfigurationProperty", "API_BlackBoxArgument", "API_BlackBoxEmpty",
        "API_CannotCheckAccessControl", "API_CannotOpenFile",
        "API_CantActivatePOAManager", "API_CantCreateClassPoa",
        "API_CantCreateLockingThread", "API_CantFindLockingThread",
        "API_CantGetClientIdent", "API_CantGetDevObjectId",
        "API_CantInstallSignal", "API_CantRetrieveClass", "API_CantRetrieveClassList",
        "API_CantStoreDeviceClass", "API_ClassNotFound",
        "API_CmdArgumentTypeNotSupported", "API_CommandNotAllowed",
        "API_CommandNotFound", "API_CorbaSysException", "API_CorruptedDatabase",
        "API_DatabaseAccess", "API_DeviceLocked", "API_DeviceNotFound",
        "API_DeviceNotLocked", "API_DeviceUnlockable", "API_DeviceUnlocked",
        "API_EventSupplierNotConstructed", "API_IncoherentDbData",
        "API_IncoherentDevData", "API_IncoherentValues", "API_IncompatibleAttrDataType",
        "API_IncompatibleCmdArgumentType", "API_InitMethodNotFound",
        "API_InitNotPublic", "API_InitThrowsException",
        "API_JavaRuntimeSecurityException", "API_MemoryAllocation",
        "API_MethodArgument", "API_MethodNotFound", "API_MissedEvents",
        "API_NotSupportedFeature", "API_NtDebugWindowError",
        "API_OverloadingNotSupported", "API_PolledDeviceNotInPoolConf",
        "API_PolledDeviceNotInPoolMap", "API_PollingThreadNotFound",
        "API_ReadOnlyMode", "API_SignalOutOfRange", "API_SystemCallFailed",
        "API_WAttrOutsideLimit", "API_WizardConfError", "API_WrongEventData",
        "API_WrongHistoryDataBuffer", "API_WrongLockingStatus", "API_ZmqInitFailed")

    for const in consts:
        setattr(mod, const, const)


def __prepare_pytango():

    if not hasattr(PyTango.constants, "API_DeviceNotFound"):
        __init_pytango_devfailed(PyTango.constants)

__prepare_pytango()



##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This is the communication macro module"""

__docformat__ = 'restructuredtext'

__all__ = ["get", "put"]

from sardana.macroserver.macro import *

import array


class put(Macro):
    """Sends a string to the communication channel"""
    param_def = [
        ['communication channel', Type.ComChannel,
            None, 'the communication channel'],
        ['data', Type.String, None, 'data to be sent']
    ]

    def run(self, comch, data):
        name = comch.getName()
        nb = comch.write(data)
        o = str(nb) + " bytes sent to " + name
        self.output(o)


class get(Macro):
    """Reads and outputs the data from the communication channel"""

    param_def = [
        ['communication channel', Type.ComChannel,
            None, 'the communication channel'],
        ['maximum length', Type.String, -1, 'maximum number of bytes to read']
    ]

    def run(self, comch, maxlen):
        name = comch.getName()
        data = comch.command_inout("read", maxlen)

        datastr = array.array('B', data).tostring()
        self.output("'" + datastr + "'")
        self.output(data)



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.tango-controls.org/static/sardana/latest/doc/html/index.html
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

__all__ = ["BasePoolTestCase"]

import logging

from sardana.pool.test import (FakePool, createPoolController, createCtrlConf,
                               createPoolCounterTimer, createPoolTriggerGate,
                               createPoolMotor, createElemConf,
                               createPoolZeroDExpChannel,
                               createPoolTwoDExpChannel,
                               createPoolPseudoCounter, createPoolPseudoMotor)


class BasePoolTestCase(object):
    """Base pool test for setting the environment."""

    POOLPATH = []
    LOGLEVEL = logging.WARNING

    def createController(self, name, klass, lib, props={}):
        c_cfg = createCtrlConf(self.pool, name, klass, lib, props)
        ctrl_obj = createPoolController(self.pool, c_cfg)
        self.ctrls[name] = ctrl_obj
        self.pool.add_element(ctrl_obj)
        return ctrl_obj

    def createCTElement(self, ctrl_obj, name, axis):
        e_cfg = createElemConf(self.pool, axis, name)
        elem_obj = createPoolCounterTimer(self.pool, ctrl_obj, e_cfg)
        ctrl_obj.add_element(elem_obj)
        # CT elements
        self.cts[name] = elem_obj
        self.pool.add_element(elem_obj)
        return elem_obj

    def createZeroDElement(self, ctrl_obj, name, axis):
        e_cfg = createElemConf(self.pool, axis, name)
        elem_obj = createPoolZeroDExpChannel(self.pool, ctrl_obj, e_cfg)
        ctrl_obj.add_element(elem_obj)
        # ZeroD elements
        self.zerods[name] = elem_obj
        self.pool.add_element(elem_obj)
        return elem_obj

    def createTwoDElement(self, ctrl_obj, name, axis):
        e_cfg = createElemConf(self.pool, axis, name)
        elem_obj = createPoolTwoDExpChannel(self.pool, ctrl_obj, e_cfg)
        ctrl_obj.add_element(elem_obj)
        # TwoD elements
        self.twods[name] = elem_obj
        self.pool.add_element(elem_obj)
        return elem_obj

    def createTGElement(self, ctrl_obj, name, axis):
        e_cfg = createElemConf(self.pool, axis, name)
        elem_obj = createPoolTriggerGate(self.pool, ctrl_obj, e_cfg)
        ctrl_obj.add_element(elem_obj)
        # TG elements
        self.tgs[name] = elem_obj
        self.pool.add_element(elem_obj)
        return elem_obj

    def createMotorElement(self, ctrl_obj, name, axis):
        e_cfg = createElemConf(self.pool, axis, name)
        elem_obj = createPoolMotor(self.pool, ctrl_obj, e_cfg)
        ctrl_obj.add_element(elem_obj)
        # MOT elements
        self.mots[name] = elem_obj
        self.pool.add_element(elem_obj)
        return elem_obj

    def createPCElement(self, ctrl_obj, name, axis, elements=[]):
        e_cfg = createElemConf(self.pool, axis, name)
        elem_obj = createPoolPseudoCounter(self.pool, ctrl_obj, e_cfg,
                                           elements)
        ctrl_obj.add_element(elem_obj)
        self.pcs[name] = elem_obj
        self.pool.add_element(elem_obj)
        return elem_obj

    def createPMElement(self, ctrl_obj, name, axis, elements=[]):
        e_cfg = createElemConf(self.pool, axis, name)
        elem_obj = createPoolPseudoMotor(self.pool, ctrl_obj, e_cfg,
                                         elements)
        ctrl_obj.add_element(elem_obj)
        self.pms[name] = elem_obj
        self.pool.add_element(elem_obj)
        return elem_obj

    def setUp(self):
        """Create a collection of controllers and elements.
        """
        self.nctctrls = self.nzerodctrls = self.ntgctrls = self.nmotctrls = 4
        # dummy controller generates an array of zeros (1024x1024) for each
        # axis, many axes may increase the memory consumption of testsuite
        self.ntwodctrls = 1
        self.nctelems = self.nzerodelems = self.ntgelems = self.nmotelems = 5
        # dummy controller generates an array of zeros (1024x1024) for each
        # axis, many axes may increase the memory consumption of testsuite
        self.ntwodelems = 1
        self.pool = FakePool(self.POOLPATH, self.LOGLEVEL)
        # Use debug mode

        self.ctrls = {}
        self.cts = {}
        self.zerods = {}
        self.twods = {}
        self.tgs = {}
        self.mots = {}
        self.pcs = {}
        self.pms = {}
        self.exp_channels = {}
        # Create nctctrls CT ctrls
        for ctrl in range(1, self.nctctrls + 1):
            name = '_test_ct_ctrl_%s' % ctrl
            ctrl_obj = self.createController(name,
                                             'DummyCounterTimerController',
                                             'DummyCounterTimerController.py')
            # use the first trigger/gate element by default
            ctrl_obj.set_ctrl_attr("synchronizer", "_test_tg_1_1")
            # Create nctelems CT elements for each ctrl
            for axis in range(1, self.nctelems + 1):
                name = '_test_ct_%s_%s' % (ctrl, axis)
                self.createCTElement(ctrl_obj, name, axis)
        self.exp_channels.update(self.cts)
        # Create nzerodctrls ZeroD ctrls
        for ctrl in range(1, self.nzerodctrls + 1):
            name = '_test_0d_ctrl_%s' % ctrl
            ctrl_obj = self.createController(name,
                                             'DummyZeroDController',
                                             'DummyZeroDController.py')
            # Create nzerodelems ZeroD elements for each ctrl
            for axis in range(1, self.nzerodelems + 1):
                name = '_test_0d_%s_%s' % (ctrl, axis)
                self.createZeroDElement(ctrl_obj, name, axis)
        self.exp_channels.update(self.zerods)
        # Create ntwodctrls TwoD ctrls
        for ctrl in range(1, self.ntwodctrls + 1):
            name = '_test_2d_ctrl_%s' % ctrl
            ctrl_obj = self.createController(name,
                                             'DummyTwoDController',
                                             'DummyTwoDController.py')
            # use the first trigger/gate element by default
            ctrl_obj.set_ctrl_attr("synchronizer", "_test_tg_1_1")
            # Create ntwodelems TwoD elements for each ctrl
            for axis in range(1, self.ntwodelems + 1):
                name = '_test_2d_%s_%s' % (ctrl, axis)
                self.createTwoDElement(ctrl_obj, name, axis)
        self.exp_channels.update(self.twods)
        # Create ntgctrls TG ctrls
        for ctrl in range(1, self.ntgctrls + 1):
            name = '_test_tg_ctrl_%s' % ctrl
            ctrl_obj = self.createController(name,
                                             'DummyTriggerGateController',
                                             'DummyTriggerGateController.py')
            # Create ntgelems TG elements for each ctrl
            for axis in range(1, self.ntgelems + 1):
                name = '_test_tg_%s_%s' % (ctrl, axis)
                self.createTGElement(ctrl_obj, name, axis)
        # Create nmotctrls MOT ctrls
        for ctrl in range(1, self.nmotctrls + 1):
            name = '_test_mot_ctrl_%s' % ctrl
            ctrl_obj = self.createController(name,
                                             'DummyMotorController',
                                             'DummyMotorController.py')
            # Create nmotelems MOT elements for each ctrl
            for axis in range(1, self.nmotelems + 1):
                name = '_test_mot_%s_%s' % (ctrl, axis)
                self.createMotorElement(ctrl_obj, name, axis)

        # Check the elements creation
        cts = len(list(self.cts.keys()))
        tgs = len(list(self.tgs.keys()))
        mots = len(list(self.mots.keys()))

        expected_cts = self.nctelems * self.nctctrls
        msg = 'Something happened during the creation of CT elements.\n' + \
              'Expected %s and there are %s, %s' % \
              (expected_cts, cts, list(self.cts.keys()))
        if cts != expected_cts:
            raise Exception(msg)
        expected_tgs = self.ntgelems * self.ntgctrls
        msg = 'Something happened during the creation of TG elements.\n' + \
              'Expected %s and there are %s, %s' % \
              (expected_tgs, tgs, list(self.tgs.keys()))
        if tgs != expected_tgs:
            raise Exception(msg)
        expected_mots = self.nmotelems * self.nmotctrls
        msg = 'Something happened during the creation of MOT elements.\n' + \
              'Expected %s and there are %s, %s' % \
              (self.nmotelems, mots, list(self.mots.keys()))
        if mots != expected_mots:
            raise Exception(msg)

    def tearDown(self):
        self.pool.cleanup()
        self.pool = None
        self.ctrls = None
        self.cts = None
        self.tgs = None



##############################################################################
##
# This file is part of Sardana
##
# http://www.tango-controls.org/static/sardana/latest/doc/html/axisex.html
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

import re
import sys
import time
import copy

import numpy
try:
    import h5py
except ImportError:
    pass

from sardana import State
from sardana.pool import AcqSynch
from sardana.pool.controller import TwoDController, Referable, \
    Type, Description, MaxDimSize, FGet, FSet, DefaultValue


def gauss(x, mean, ymax, fwhm, yoffset=0):
    return yoffset + ymax * numpy.power(2, -4 * ((x - mean) / fwhm)**2)


def generate_img(x_size, y_size, amplitude):
    x = numpy.linspace(-10, 10, x_size)
    y = numpy.linspace(-10, 10, y_size)
    x, y = numpy.meshgrid(x, y)
    img = (gauss(x, 0, amplitude, 4) * gauss(y, 0, amplitude, 4))
    return img


def generate_ref(pattern, idx):
    if pattern is None or pattern == "":
        pattern = "h5file:///tmp/dummy2d_default_{index}.h5"
    msg = None
    try:
        uri = pattern.format(index=idx)
    except Exception:
        uri = pattern
        msg = ("Not able to format value reference template "
               "with index. Trying to use directly the template...")
    match_res = re.match(
        r"(?P<scheme>h5file)://(?P<path>\S+)::(?P<dataset>\S+)",
        uri)
    if match_res is None:
        match_res = re.match(
            r"(?P<scheme>(h5file|file))://(?P<path>\S+)",
            uri)
    if match_res is None:
        raise Exception("invalid value reference template")
    scheme = match_res.group("scheme")
    path = match_res.group("path")
    try:
        dataset_name = match_res.group("dataset")
    except IndexError:
        if scheme == "h5file":
            dataset_name = "dataset"
        else:
            dataset_name = None
    return scheme, path, dataset_name, msg


def save_img(img, path, dataset_name):
    msg = None
    if "h5py" not in sys.modules:
        msg = "Not able to store h5 file (h5py is not available)"
    try:
        h5f = h5py.File(path, "w")
        h5f.create_dataset(dataset_name, data=img)
    except Exception:
        msg = "Not able to store h5 file."
    return msg


class Channel:

    def __init__(self, idx):
        self.idx = idx            # 1 based index
        self.value = []
        self.value_ref = None
        self.is_counting = False
        self.acq_idx = 0
        self.buffer_values = []
        self.buffer_value_refs = []
        self.amplitude = BaseValue('1.0')
        self.saving_enabled = False
        self.value_ref_pattern = "h5file:///tmp/dummy2d_default_{index}.h5"
        self.value_ref_enabled = False


class BaseValue(object):

    def __init__(self, value):
        self.raw_value = value
        self.init()

    def init(self):
        self.value = float(self.raw_value)

    def get(self):
        return self.value

    def get_value_name(self):
        return self.raw_value


class TangoValue(BaseValue):

    def init(self):
        import PyTango
        self.attr_proxy = PyTango.AttributeProxy(self.raw_value)

    def get(self):
        return self.attr_proxy.read().value


class BasicDummyTwoDController(TwoDController):
    """This class represents a basic, dummy Sardana TwoD controller."""

    gender = "Simulation"
    model = "Basic"
    organization = "Sardana team"

    MaxDevice = 1024

    BufferSize = 1024, 1024

    default_latency_time = 0.0

    ctrl_attributes = {
        "Synchronizer": {
                Type: str,
                Description: ("Hardware (external) emulated synchronizer. "
                              "Can be any of dummy trigger/gate elements "
                              "from the same pool.")
            },
    }

    axis_attributes = {
        'Amplitude': {
            Type: str,
            FGet: 'getAmplitude',
            FSet: 'setAmplitude',
            Description: ("Amplitude. Maybe a number or a tango attribute "
                          "(must start with tango://)"),
            DefaultValue: '1.0'}
    }

    def __init__(self, inst, props, *args, **kwargs):
        TwoDController.__init__(self, inst, props, *args, **kwargs)
        self.channels = self.MaxDevice * [None, ]
        self.start_time = None
        self.integ_time = None
        self.repetitions = None
        self.latency_time = None
        self.acq_cycle_time = None  # integ_time + latency_time
        self.estimated_duration = None
        self.start_idx = None
        self._synchronization = AcqSynch.SoftwareTrigger
        self.read_channels = {}
        self.counting_channels = {}
        # name of synchronizer element
        self._synchronizer = None
        # synchronizer element (core)
        self.__synchronizer_obj = None
        # flag whether the controller was armed for hardware synchronization
        self._armed = False

    def GetAxisAttributes(self, axis):
        # the default max shape for 'value' is (16*1024,).
        # We don't need so much so we set it to BufferSize
        attrs = super(BasicDummyTwoDController, self).GetAxisAttributes(axis)
        attrs['Value'][MaxDimSize] = self.BufferSize
        return attrs

    def AddDevice(self, axis):
        idx = axis - 1
        self.channels[idx] = channel = Channel(axis)
        channel.value = numpy.zeros(self.BufferSize, dtype=numpy.float64)

    def DeleteDevice(self, axis):
        idx = axis - 1
        self.channels[idx] = None

    def PrepareOne(self, axis, value, repetitions, latency, nb_starts):
        self.start_idx = -1

    def LoadOne(self, axis, integ_time, repetitions, latency_time):
        self.integ_time = integ_time
        self.repetitions = repetitions
        self.latency_time = latency_time
        self.acq_cycle_time = acq_cycle_time = integ_time + latency_time
        self.estimated_duration = acq_cycle_time * repetitions - latency_time

    def PreStartAll(self):
        self.counting_channels = {}
        self.read_channels = {}
        self.start_idx += 1

    def PreStartOne(self, axis, value):
        idx = axis - 1
        channel = self.channels[idx]
        channel.value = None
        channel.acq_idx = 0
        channel.buffer_values = []
        self.counting_channels[axis] = channel
        self.read_channels[axis] = channel
        return True

    def StartOne(self, axis, value):
        if self._synchronization in (AcqSynch.SoftwareStart,
                                     AcqSynch.SoftwareTrigger):
            self.counting_channels[axis].is_counting = True

    def StartAll(self):
        if self._synchronization in (AcqSynch.HardwareStart,
                                     AcqSynch.HardwareTrigger,
                                     AcqSynch.HardwareGate):
            self._connect_hardware_synchronization()
            self._armed = True
        else:
            self.start_time = time.time()

    def _updateChannelState(self, axis, elapsed_time):
        if self._synchronization == AcqSynch.SoftwareTrigger:
            if self.integ_time is not None:
                # counting in time
                if elapsed_time >= self.integ_time:
                    self._finish(elapsed_time)
        elif self._synchronization in (AcqSynch.HardwareTrigger,
                                       AcqSynch.HardwareGate,
                                       AcqSynch.HardwareStart,
                                       AcqSynch.SoftwareStart):
            if self.integ_time is not None:
                # counting in time
                if elapsed_time > self.estimated_duration:
                    self._finish(elapsed_time)

    def StateOne(self, axis):
        idx = axis - 1
        sta = State.On
        status = "Stopped"
        if self._armed:
            sta = State.Moving
            status = "Armed"
        elif axis in self.counting_channels and self.start_time is not None:
            channel = self.channels[idx]
            now = time.time()
            elapsed_time = now - self.start_time
            self._updateChannelState(axis, elapsed_time)
            if channel.is_counting:
                sta = State.Moving
                status = "Acquiring"
                # TODO: do it only once at the end
                self._updateChannelValue(axis, elapsed_time)
        return sta, status

    def _updateChannelValue(self, axis, elapsed_time):
        channel = self.channels[axis - 1]
        if channel.acq_idx == self.repetitions:
            return
        x_size = self.BufferSize[0]
        y_size = self.BufferSize[1]
        amplitude = axis * self.integ_time * channel.amplitude.get()
        img = generate_img(x_size, y_size, amplitude)
        if self._synchronization == AcqSynch.SoftwareTrigger:
            channel.value = img
            channel.acq_idx += 1
        elif self._synchronization in (AcqSynch.HardwareTrigger,
                                       AcqSynch.HardwareGate,
                                       AcqSynch.HardwareStart,
                                       AcqSynch.SoftwareStart):
            acq_cycle_time = self.acq_cycle_time
            nb_elapsed_acq, resting = divmod(elapsed_time, acq_cycle_time)
            nb_elapsed_acq = int(nb_elapsed_acq)
            # do not wait the last latency_time
            if (nb_elapsed_acq == self.repetitions - 1
                    and resting > self.integ_time):
                nb_elapsed_acq += 1
            if nb_elapsed_acq > self.repetitions:
                nb_elapsed_acq = self.repetitions
            nb_new_acq = nb_elapsed_acq - channel.acq_idx
            if nb_new_acq == 0:
                return
            channel.buffer_values.extend([img] * nb_new_acq)
            channel.acq_idx += nb_new_acq

    def ReadOne(self, axis):
        self._log.debug('ReadOne(%d): entering...' % axis)
        try:
            channel = self.read_channels[axis]
        except KeyError:
            # TODO: After SEP17 this won't be necessary anymore.
            msg = "no acquisition done on axis {0} so far".format(axis)
            raise RuntimeError(msg)
        ret = None
        if self._synchronization in (AcqSynch.HardwareTrigger,
                                     AcqSynch.HardwareGate,
                                     AcqSynch.HardwareStart,
                                     AcqSynch.SoftwareStart):
            values = copy.deepcopy(channel.buffer_values)
            channel.buffer_values.__init__()
            ret = values
        elif self._synchronization == AcqSynch.SoftwareTrigger:
            ret = channel.value
        self._log.debug('ReadOne(%d): returning %s' % (axis, repr(ret)))
        return ret

    def _finish(self, elapsed_time, axis=None):
        if axis is None:
            for axis, channel in list(self.counting_channels.items()):
                channel.is_counting = False
                self._updateChannelValue(axis, elapsed_time)
        elif axis in self.counting_channels:
            channel = self.counting_channels[axis]
            channel.is_counting = False
            self._updateChannelValue(axis, elapsed_time)
            self.counting_channels.pop(axis)
        if self._synchronization in (AcqSynch.HardwareStart,
                                     AcqSynch.HardwareTrigger,
                                     AcqSynch.HardwareGate):
            self._disconnect_hardware_synchronization()
            self._armed = False
        self.start_time = None

    def AbortOne(self, axis):
        if axis not in self.counting_channels:
            return
        now = time.time()
        if self.start_time is not None:
            elapsed_time = now - self.start_time
        else:
            elapsed_time = 0
        self._finish(elapsed_time, axis)

    def getAmplitude(self, axis):
        idx = axis - 1
        channel = self.channels[idx]
        return channel.amplitude.get_value_name()

    def setAmplitude(self, axis, value):
        idx = axis - 1
        channel = self.channels[idx]

        klass = BaseValue
        if value.startswith("tango://"):
            klass = TangoValue
        channel.amplitude = klass(value)

    def GetCtrlPar(self, par):
        if par == "synchronization":
            return self._synchronization
        elif par == "latency_time":
            return self.default_latency_time

    def SetCtrlPar(self, par, value):
        if par == "synchronization":
            self._synchronization = value

    def getSynchronizer(self):
        if self._synchronizer is None:
            return "None"
        else:
            # get synchronizer object to only check it exists
            self._synchronizer_obj
            return self._synchronizer

    def setSynchronizer(self, synchronizer):
        if synchronizer == "None":
            synchronizer = None
        self._synchronizer = synchronizer
        self.__synchronizer_obj = None  # invalidate cache

    @property
    def _synchronizer_obj(self):
        """Get synchronizer object with cache mechanism.

        If synchronizer object is not cached ("""
        if self.__synchronizer_obj is not None:
            return self.__synchronizer_obj
        synchronizer = self._synchronizer
        if synchronizer is None:
            msg = "Hardware (external) emulated synchronizer is not set"
            raise ValueError(msg)
        # getting pool (core) element - hack
        pool_ctrl = self._getPoolController()
        pool = pool_ctrl.pool
        try:
            synchronizer_obj = pool.get_element_by_name(synchronizer)
        except Exception:
            try:
                synchronizer_obj = pool.get_element_by_full_name(synchronizer)
            except Exception:
                msg = "Unknown synchronizer {0}".format(synchronizer)
                raise ValueError(msg)
        self.__synchronizer_obj = synchronizer_obj
        return synchronizer_obj

    def _connect_hardware_synchronization(self):
        # obtain dummy trigger/gate controller (plugin) instance - hack
        tg_ctrl = self._synchronizer_obj.controller.ctrl
        idx = self._synchronizer_obj.axis - 1
        func_generator = tg_ctrl.tg[idx]
        func_generator.add_listener(self)

    def _disconnect_hardware_synchronization(self):
        # obtain dummy trigger/gate controller (plugin) instance - hack
        tg_ctrl = self._synchronizer_obj.controller.ctrl
        idx = self._synchronizer_obj.axis - 1
        func_generator = tg_ctrl.tg[idx]
        func_generator.remove_listener(self)

    def event_received(self, src, type_, value):
        """Callback for dummy trigger/gate function generator events
        e.g. start, active passive
        """
        # for the moment only react on first trigger
        if type_.name.lower() == "active" and value == 0:
            self._armed = False
            for axis, channel in self.counting_channels.items():
                channel.is_counting = True
            self.start_time = time.time()


class DummyTwoDController(BasicDummyTwoDController, Referable):
    """This class is the Tango Sardana TwoDController controller for tests"""

    model = "Best"

    axis_attributes = dict(BasicDummyTwoDController.axis_attributes)
    axis_attributes.update(
        {
             "SavingEnabled": {
                Type: bool,
                FGet: "isSavingEnabled",
                FSet: "setSavingEnabled",
                Description: ("Enable/disable saving of images in HDF5 files."
                              " Use with care in high demanding (fast)"
                              " acquisitions. Trying to save at high rate may"
                              " hang the acquisition process.")
             }
        }
    )

    def __init__(self, inst, props, *args, **kwargs):
        BasicDummyTwoDController.__init__(self, inst, props, *args, **kwargs)

    def PrepareOne(self, axis, value, repetitions, latency, nb_starts):
        BasicDummyTwoDController.PrepareOne(self, axis, value, repetitions,
                                            latency, nb_starts)
        idx = axis - 1
        channel = self.channels[idx]
        value_ref_pattern = channel.value_ref_pattern
        saving_enabled = channel.saving_enabled
        # just validate if the pattern is correct, index does not matter
        # that's why 0 is used
        scheme, path, dataset_name, msg = generate_ref(value_ref_pattern,
                                                       0)
        if scheme != "h5file" and saving_enabled:
            raise Exception("only h5file saving is supported")

    def _updateChannelValue(self, axis, elapsed_time):
        channel = self.channels[axis - 1]
        if channel.acq_idx == self.repetitions:
            return
        x_size = self.BufferSize[0]
        y_size = self.BufferSize[1]
        amplitude = axis * self.integ_time * channel.amplitude.get()
        img = generate_img(x_size, y_size, amplitude)
        if self._synchronization == AcqSynch.SoftwareTrigger:
            channel.value = img
            if channel.value_ref_enabled:
                img_idx = self.start_idx * self.repetitions + channel.acq_idx
                value_ref_pattern = channel.value_ref_pattern
                scheme, path, dataset_name, msg = generate_ref(
                    value_ref_pattern, img_idx)
                if msg is not None:
                    self._log.warning(msg)
                value_ref = scheme + "://" + path
                if channel.saving_enabled:
                    msg = save_img(img, path, dataset_name)
                    if msg is not None:
                        self._log.warning(msg)
                    else:
                        # we succeeded to save in HDF5
                        value_ref = value_ref + "::" + dataset_name
                channel.value_ref = value_ref
            channel.acq_idx += 1
        elif self._synchronization in (AcqSynch.HardwareTrigger,
                                       AcqSynch.HardwareGate,
                                       AcqSynch.HardwareStart,
                                       AcqSynch.SoftwareStart):
            acq_cycle_time = self.acq_cycle_time
            nb_elapsed_acq, resting = divmod(elapsed_time, acq_cycle_time)
            nb_elapsed_acq = int(nb_elapsed_acq)
            # do not wait the last latency_time
            if (nb_elapsed_acq == self.repetitions - 1
                    and resting > self.integ_time):
                nb_elapsed_acq += 1
            if nb_elapsed_acq > self.repetitions:
                nb_elapsed_acq = self.repetitions
            nb_new_acq = nb_elapsed_acq - channel.acq_idx
            if nb_new_acq == 0:
                return
            channel.buffer_values.extend([img] * nb_new_acq)
            if channel.value_ref_enabled:
                start = self.start_idx * self.repetitions + channel.acq_idx
                for img_idx in range(start, start + nb_new_acq):
                    value_ref_pattern = channel.value_ref_pattern
                    scheme, path, dataset_name, msg = generate_ref(
                        value_ref_pattern, img_idx)
                    if msg is not None:
                        self._log.warning(msg)
                    value_ref = scheme + "://" + path
                    if channel.saving_enabled:
                        msg = save_img(img, path, dataset_name)
                        if msg is not None:
                            self._log.warning(msg)
                        else:
                            # we succeeded to save in HDF5
                            value_ref = value_ref + "::" + dataset_name
                    channel.buffer_value_refs.append(value_ref)
            channel.acq_idx += nb_new_acq

    def RefOne(self, axis):
        self._log.debug("RefOne(%s)", axis)
        channel = self.read_channels[axis]
        ret = None
        if self._synchronization in (AcqSynch.HardwareTrigger,
                                     AcqSynch.HardwareGate,
                                     AcqSynch.HardwareStart,
                                     AcqSynch.SoftwareStart,):
            value_refs = copy.deepcopy(channel.buffer_value_refs)
            channel.buffer_value_refs.__init__()
            ret = value_refs
        elif self._synchronization == AcqSynch.SoftwareTrigger:
            ret = channel.value_ref
        self._log.debug('RefOne(%d): returning %s' % (axis, repr(ret)))
        return ret

    def SetAxisPar(self, axis, parameter, value):
        idx = axis - 1
        channel = self.channels[idx]
        if parameter == "value_ref_pattern":
            channel.value_ref_pattern = value
        elif parameter == "value_ref_enabled":
            channel.value_ref_enabled = value

    def isSavingEnabled(self, axis):
        idx = axis - 1
        channel = self.channels[idx]
        return channel.saving_enabled

    def setSavingEnabled(self, axis, value):
        idx = axis - 1
        channel = self.channels[idx]
        channel.saving_enabled = value



import PyTango
import socket
import MotorController

nb_call = 0


class IcePapController(MotorController.MotorController):

    MaxDevice = 4

    def __init__(self, inst, props):
        print "PYTHON -> IcePapController ctor for instance", inst
        MotorController.MotorController.__init__(self, inst, props)
        self.nb_call = 0
        self.socket_connected = False
        self.db = PyTango.Database()
        self.ct_name = "IcePapController/" + self.inst_name

#
# Get controller properties
#

        prop_list = ['host', 'port', 'timeout']
        prop = self.db.get_property(self.ct_name, prop_list)

        if len(prop["host"]) != 0:
            self.host = prop["host"][0]
        else:
            print "Property host not defined for controller", self.ct_name
            self.host = "nada"

        if len(prop["port"]) != 0:
            self.port = int(prop["port"][0])
        else:
            print "Property port not defined for controller", self.ct_name
            self.port = 0

        if len(prop["timeout"]) != 0:
            self.timeout = int(prop["timeout"][0])
        else:
            print "Property timeout not defined for controller", self.ct_name
            self.timeout = 3

 #
 # Connect to the icepap
 #

        print "PYTHON -> IcePap on", self.host, " and port", self.port, " with timeout = ", self.timeout

        global nb_call

        if nb_call <= 1:
            nb_call = nb_call + 1
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.sock.settimeout(self.timeout)
            self.sock.connect((self.host, self.port))
            self.socket_connected = True

        print "PYTHON -> Connected to", self.host, " on port", self.port

#
# Check that the Icepap is OK
#

#        ans = self.IceWriteRead("?ID")

    def AddDevice(self, axis):
        print "PYTHON -> IcePapController/", self.inst_name, ": In AddDevice method for axis", axis
#        raise RuntimeError,"Hola la la"

    def DeleteDevice(self, axis):
        print "PYTHON -> IcePapController/", self.inst_name, ": In DeleteDevice method for axis", axis

    def StateOne(self, axis):
        print "PYTHON -> IcePapController/", self.inst_name, ": In StateOne method for axis", axis
        tup = (PyTango.DevState.ON, 0)
        return tup

    def PreReadAll(self):
        print "PYTHON -> IcePapController/", self.inst_name, ": In PreReadAll method"

    def PreReadOne(self, axis):
        print "PYTHON -> IcePapController/", self.inst_name, ": In PreReadOne method for axis", axis

    def ReadAll(self):
        print "PYTHON -> IcePapController/", self.inst_name, ": In ReadAll method"

    def ReadOne(self, axis):
        print "PYTHON -> IcePapController/", self.inst_name, ": In ReadOne method for axis", axis
        return 123

    def PreStartAll(self):
        print "PYTHON -> IcePapController/", self.inst_name, ": In PreStartAll method"

    def PreStartOne(self, axis, pos):
        print "PYTHON -> IcePapController/", self.inst_name, ": In PreStartOne method for axis", axis, " with pos", pos
        return True

    def StartOne(self, axis, pos):
        print "PYTHON -> IcePapController/", self.inst_name, ": In StartOne method for axis", axis, " with pos", pos

    def StartAll(self):
        print "PYTHON -> IcePapController/", self.inst_name, ": In StartAll method"

    def SetPar(self, axis, name, value):
        print "PYTHON -> IcePapController/", self.inst_name, ": In SetPar method for axis", axis, " name=", name, " value=", value

    def GetPar(self, axis, name):
        print "PYTHON -> IcePapController/", self.inst_name, ": In GetPar method for axis", axis, " name=", name
        return 12.34

    def IceWrite(self, data):
        data = data + "\n"
        byteSent = self.sock.send(data)
        print "PYTHON -> Sent", byteSent, "bytes to icepap"

    def IceWriteRead(self, data):
        self.IceWrite(data)
        byteReceived = self.sock.recv(1024)
        print "PYTHON -> Icepap answered:", byteReceived
        return byteReceived

    def IceResetFifo(self):
        self.IceWrite("fiforst")

    def IceCheckError(self, ice_answer):
        if (ice_answer.find("ERROR") != -1):
            new_ans = self.IceWriteRead("?ERR 1")
            print "Error from IcePap =", new_ans

    def __del__(self):
        print "PYTHON -> IcePapController/", self.inst_name, ": Aarrrrrg, I am dying"

#
# Reset IcePap FIFO
#

        if (self.socket_connected):
            print "PYTHON -> Closing connection"
            self.IceResetFifo()
            self.sock.close()


if __name__ == "__main__":
    obj = IcePapController('test')
#    obj.AddDevice(2)
#    obj.DeleteDevice(2)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""Parameter parsing"""

__all__ = ['AbstractParam', 'Param', 'ParamRepeat', 'ParamList']

import sys


class AbstractParam:

    def __init__(self, name=None, desc=None):
        self.name = name
        self.desc = desc

    def getParam(self, idx=0):
        return self

    def getParamStr(self):
        return "<" + self.name + ">"

    def getParamDescr(self):
        str = self.name + " (" + self.type + ") - " + self.desc
        if self.defvalue != 'None':
            str = str + "\nDefault value: " + self.defvalue
        return str

    def getParamCount(self):
        return 1

    def formatParamValue(self, value):
        return value

    def __str__(self):
        return "%s(%s)" % (self.__class__.__name__, self.name)


class Param(AbstractParam):

    def __init__(self, name=None, type_name=None, desc=None, defvalue=None,
                 from_array=None):
        AbstractParam.__init__(self, name=name, desc=desc)
        self.type = type_name
        self.defvalue = defvalue

        if from_array is not None:
            for key in ['name', 'type', 'desc', 'defvalue']:
                setattr(self, key, from_array.read())

    def formatParamValue(self, value):
        ret = value
        if self.type == "File":
            f = open(ret, 'r')
            ret = f.read()
            f.close()
        return ret


class ParamRepeat(AbstractParam):

    def __init__(self, name=None, desc=None, opts=None, param_def=None,
                 from_array=None):
        AbstractParam.__init__(self, name=name, desc=desc)
        self.opts = opts
        self.param_def = param_def

        if from_array is not None:
            self.name = from_array.read()
            type_name = from_array.read()
            if type_name != 'ParamRepeat':
                msg = 'Expecting "ParamRepeat" type, got ' + type_name
                raise ValueError(msg)
            self.desc = from_array.read()
            opt_str = from_array.read()
            opt_list = opt_str.split(', ')
            opts = {}
            for opt_str in opt_list:
                name, val = opt_str.split('=')
                if val == repr(None):
                    val = None
                else:
                    try:
                        val = int(val)
                    except:
                        pass
                opts[name] = val
            self.opts = opts
            self.param_def = ParamList(from_array=from_array)

    def getParam(self, idx=0):
        idx = idx % len(self.param_def.pars)
        return self.param_def.getParam(idx)

    def getParamStr(self):
        return "[" + self.param_def.getParamStr() + " ]"

    def getParamDescr(self):
        return self.param_def.getParamDescr()

    def getParamCount(self):
        return self.param_def.getParamCount()

    def formatParamValue(self, value):
        return self.param_def.formatParamValue(value)


class ParamList(AbstractParam):

    def __init__(self, from_array):
        AbstractParam.__init__(self, name="ParamList",
                               desc="List of parameters")

        self.pars = []
        self.has_param_repeat = False

        npars = int(from_array.read())
        for i in range(npars):
            if from_array[1] == 'ParamRepeat':
                par = ParamRepeat(from_array=from_array)
                self.has_param_repeat = True
            else:
                par = Param(from_array=from_array)
            self.pars.append(par)

    def getParam(self, idx=0):
        par_nb = len(self.pars)
        if idx < par_nb:
            par = idx
            return self.pars[idx].getParam(0)
        else:
            for i in range(par_nb):
                local_idx = idx - i
                local_nb = self.pars[i].getParamCount()
                if local_idx < local_nb:
                    return self.pars[i].getParam(local_idx)

    def getParamStr(self):
        str = ""
        for par in self.pars:
            str = str + " " + par.getParamStr()
        return str

    def getParamDescr(self):
        str = ""
        for par in self.pars:
            str = str + "\n" + par.getParamDescr()
        return str

    def getParamCount(self):
        nb = 0
        for par in self.pars:
            local_nb = par.getParamCount()
            if local_nb == sys.maxsize:
                return sys.maxsize
            nb += local_nb
        return nb

    def formatParamValue(self, value):

        # for now we are not able to handle ParamRepeat with special value
        # format
        if self.has_param_repeat:
            return value

        for i, v in enumerate(value):
            value[i] = self.pars[i].formatParamValue(v)
        return value



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""This module contains a taurus ShowScanOnline widget."""

__all__ = ["ShowScanOnline"]

import click

from taurus.qt.qtgui.taurusgui import TaurusGui
from sardana.taurus.qt.qtgui.macrolistener import (DynamicPlotManager,
                                                   assertPlotAvailability)


class ShowScanOnline(DynamicPlotManager):

    def __init__(self, parent):
        DynamicPlotManager.__init__(self, parent)

    def onExpConfChanged(self, expconf):
        DynamicPlotManager.onExpConfChanged(self, expconf)
        activeMntGrp = expconf['ActiveMntGrp']
        msg = 'Plotting scans for "%s" Measurement Group' % activeMntGrp
        self.parent().newShortMessage.emit(msg)

    def createPanel(self, widget, name, **kwargs):
        ''' Reimplemented from :class:`DynamicPlotManager` to delegate panel
        management to the parent widget (a TaurusGui)'''
        mainwindow = self.parent()
        return mainwindow.createPanel(widget, name, **kwargs)

    def getPanelWidget(self, name):
        ''' Reimplemented from :class:`DynamicPlotManager` to delegate panel
        management to the parent widget (a TaurusGui)'''
        mainwindow = self.parent()
        return mainwindow.getPanel(name).widget()

    def removePanel(self, name):
        ''' Reimplemented from :class:`DynamicPlotManager` to delegate panel
        management to the parent widget (a TaurusGui)'''
        mainwindow = self.parent()
        mainwindow.removePanel(name)

    def removeTemporaryPanels(self, names=None):
        '''Remove temporary panels managed by this widget'''
        # for now, the only temporary panels are the plots
        DynamicPlotManager.removePanels(self, names=names)


class TaurusGuiLite(TaurusGui):
    HEARTBEAT = None
    FILE_MENU_ENABLED = False
    VIEW_MENU_ENABLED = False
    TAURUS_MENU_ENABLED = False
    TOOLS_MENU_ENABLED = False
    PANELS_MENU_ENABLED = False
    HELP_MENU_ENABLED = False
    FULLSCREEN_TOOLBAR_ENABLED = False
    APPLETS_TOOLBAR_ENABLED = False
    QUICK_ACCESS_TOOLBAR_ENABLED = False
    USER_PERSPECTIVES_ENABLED = False
    LOGGER_WIDGET_ENABLED = False
    SPLASH_LOGO_NAME = None


@click.command()
@click.option('--group', default='x-axis',
              type=click.Choice(['single', 'x-axis']),
              help='group curves')
@click.option('--taurus-log-level',
              type=click.Choice(['critical', 'error', 'warning', 'info',
                                 'debug', 'trace']),
              default='error', show_default=True,
              help='Show only logs with priority LEVEL or above')
@click.argument('door')
def main(group, taurus_log_level, door):
    import taurus
    taurus.setLogLevel(getattr(taurus, taurus_log_level.capitalize()))

    from taurus.qt.qtgui.application import TaurusApplication

    app = TaurusApplication(app_name='Showscan Online', org_domain="Sardana",
                            org_name="Tango communinity", cmd_line_parser=None)

    assertPlotAvailability()

    gui = TaurusGuiLite()

    widget = ShowScanOnline(gui)
    widget.setModel(door)
    widget.setGroupMode(group)
    gui.show()
    return app.exec_()


if __name__ == "__main__":
    main()



#!/usr/bin/env python

##############################################################################
##
# This file is part of Sardana
##
# http://www.sardana-controls.org/
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Sardana is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Sardana is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Sardana.  If not, see <http://www.gnu.org/licenses/>.
##
##############################################################################

"""
This module provides a button for executing macros
"""

__all__ = ['MacroButton']

import functools
import shlex
import re

import PyTango

import taurus
from taurus.core import TaurusEventType, TaurusDevice
from taurus.external.qt import Qt, compat
from taurus.qt.qtgui.container import TaurusWidget
from taurus.qt.qtgui.base import TaurusBaseWidget
from taurus.qt.qtgui.button import TaurusCommandButton
from taurus.qt.qtgui.dialog import ProtectTaurusMessageBox
from taurus.core.util.colors import DEVICE_STATE_PALETTE
from taurus.qt.qtgui.util.ui import UILoadable

from sardana.util.parser import ParamParser


class DoorStateListener(Qt.QObject):
    '''A listener of Change and periodic events from a Door State attribute.
    It converts the received Tango events and emits a Qt signal
    '''

    __pyqtSignals__ = ["doorStateChanged"]

    doorStateChanged = Qt.pyqtSignal(compat.PY_OBJECT)

    def eventReceived(self, evt_src, evt_type, evt_value):
        if evt_type not in (TaurusEventType.Change, TaurusEventType.Periodic):
            return
        door_state = evt_value.value
        self.doorStateChanged.emit(door_state)


@UILoadable(with_ui='ui')
class MacroButton(TaurusWidget):
    ''' A button to execute/pause/stop macros. The model must be a valid door.

    .. todo:: Not implemented but will be needed: set an icon

    .. todo:: It may be useful to have all the streams from qdoor available
             somehow (right-click?)
    '''

    __pyqtSignals__ = ['statusUpdated', 'resultUpdated']

    statusUpdated = Qt.pyqtSignal(compat.PY_OBJECT)
    resultUpdated = Qt.pyqtSignal(compat.PY_OBJECT)

    def __init__(self, parent=None, designMode=False):
        TaurusWidget.__init__(self, parent, designMode)
        self.loadUi()
        self.door = None
        self.door_state_listener = None
        self.macro_name = ''
        self.macro_args = []
        self.macro_id = None
        self.running_macro = None

        self.ui.progress.setValue(0)

        self.ui.button.setCheckable(True)
        self.ui.button.clicked.connect(self._onButtonClicked)

    # Override default implementation of handleEvent from TaurusWidget
    # in order to avoid button's text being lost on the MS restart.
    # More detais in #293 and taurus-org/taurus#635
    def handleEvent(self, evt_src, evt_type, evt_value):
        pass

    def showProgress(self, visible):
        '''Set whether the progress bar is shown

        :param visible: (bool) If True, the progress bar is shown. Otherwise it
                        is hidden'''
        self.ui.progress.setVisible(visible)

    def setModel(self, model):
        '''
        reimplemented from :class:`TaurusWidget`. A door device name is
        expected as the model
        '''
        TaurusWidget.setModel(self, model)
        if self.door is not None:
            self.door.macroStatusUpdated.disconnect(self._statusUpdated)
            self.door.resultUpdated.disconnect(self._resultUpdated)

            # disable management of Door Tango States
            self.door.getAttribute('State').removeListener(
                self.door_state_listener)
            self.door_state_listener.doorStateChanged.disconnect(
                self._doorStateChanged)
            self.door_state_listener = None

        try:
            self.door = taurus.Device(model)
        except:
            return

        self.door.macroStatusUpdated.connect(self._statusUpdated)
        self.door.resultUpdated.connect(self._resultUpdated)

        # Manage Door Tango States
        self.door_state_listener = DoorStateListener()
        self.door_state_listener.doorStateChanged.connect(
            self._doorStateChanged)
        self.door.getAttribute('State').addListener(self.door_state_listener)

    def _doorStateChanged(self, state):
        '''slot called on door state changes'''
        color = '#' + DEVICE_STATE_PALETTE.hex(state)
        stylesheet = 'QFrame{border: 4px solid %s;}' % color
        self.ui.frame.setStyleSheet(stylesheet)

        # In case state is not ON, and macro not triggered by the button,
        # disable it
        door_available = True
        if state not in [PyTango.DevState.ON, PyTango.DevState.ALARM] and not self.ui.button.isChecked():
            door_available = False

        self.ui.button.setEnabled(door_available)
        self.ui.progress.setEnabled(door_available)

    def _statusUpdated(self, *args):
        '''slot called on status changes'''
        # SHOULD SEE THE DOCUMENTATION ABOUT THE ARGS AND ALSO THE STATUS STATE MACHINE
        # ARGS FORMAT IS (GUESSING WITH PRINT STATEMENTS)
        # e.g. ((<sardana.taurus.core.tango.sardana.macro.Macro object at 0x7f29300bc210>, [{u'step': 100.0, u'state': u'stop', u'range': [0.0, 100.0], u'id': u'b226f5e8-c807-11e0-8abe-001d0969db5b'}]),)
        # ( (MacroObj, [status_dict, .?.]), .?.)

        # QUESTIONS: THIS MACRO OBJECT HAS ALOS STEP, RANGE, ...
        # AND ALSO THE STATUS DICT... WHICH SHOULD I USE?

        first_tuple = args[0]
        self.running_macro = first_tuple[0]

        status_dict = first_tuple[1][0]
        # KEYS RECEIVED FROM A 'SCAN' MACRO AND A 'TWICE' MACRO: IS IT GENERAL
        # ?!?!?!
        macro_id = status_dict['id']
        # if macro id is unknown ignoring this signal
        if macro_id is None:
            return
        # check if we have launch this macro, otherwise ignore the signal
        if macro_id != str(self.macro_id):
            return
        state = status_dict['state']
        step = status_dict['step']
        step_range = status_dict['range']

        # Update progress bar
        self.ui.progress.setMinimum(step_range[0])
        self.ui.progress.setMaximum(step_range[1])
        self.ui.progress.setValue(step)

        if state in ['stop', 'abort', 'finish', 'alarm']:
            self.ui.button.setChecked(False)

        self.statusUpdated.emit(status_dict)

    def _resultUpdated(self, *args):
        '''slot called on result changes'''
        # ARGS APPEAR TO BE EMPTY... SHOULD THEY CONTAIN THE RESULT ?!?!?!
        # I have to rely on the 'macro object' received in the last status
        # update
        if self.running_macro is None:
            return
        result = self.running_macro.getResult()
        self.resultUpdated.emit(result)

    def setText(self, text):
        '''set the button text

        :param text: (str) text for the button
        '''
        self.setButtonText(text)

    def setButtonText(self, text):
        '''same as :meth:`setText`
        '''
        # SHOULD ALSO BE POSSIBLE TO SET AN ICON
        self.ui.button.setText("Run/Stop:\n" + text)

    def setMacroName(self, name):
        '''set the name of the macro to be executed

        :param name: (str) text for the button
        '''
        self.macro_name = str(name)
        # update tooltip
        self.setToolTip(self.macro_name + ' ' + ' '.join(self.macro_args))

    def updateMacroArgument(self, index, value):
        '''change a given argument

        :param index: (int) positional index for this argument
        :param value: value for this argument
        '''
        # make sure that the macro_args is at least as long as index
        while len(self.macro_args) < index + 1:
            self.macro_args.append('')
        # some signals may come with other than string argumenst e.g. int
        # so convert them to string
        value = str(value)
        # string arguments may contain spaces so encapsulate them in quotes
        if re.search('\s', value):
            value = '"{0}"'.format(value)
        # update the given argument
        self.macro_args[index] = value
        # update tooltip
        self.setToolTip(self.macro_name + ' ' + ' '.join(self.macro_args))

    @staticmethod
    def __isSignal(obj):
        if not hasattr(obj, 'emit'):
            return False
        if not hasattr(obj, 'connect'):
            return False
        if not hasattr(obj, 'disconnect'):
            return False
        return True

    def _onButtonClicked(self):
        if self.ui.button.isChecked():
            self.runMacro()
        else:
            self.stop()

    @ProtectTaurusMessageBox(msg='Error while executing the macro.')
    def runMacro(self):
        '''execute the macro with the current arguments'''
        if self.door is None:
            return
        param_defs = self.door.macro_server.getMacroInfoObj(
            self.macro_name).parameters
        parser = ParamParser(param_defs)
        parameters = parser.parse(" ".join(self.macro_args))
        try:
            self.door.runMacro(self.macro_name, parameters)
            sec_xml = self.door.getRunningXML()
            # get the id of the current running macro
            self.macro_id = sec_xml[0].get("id")
        except Exception as e:
            self.ui.button.setChecked(False)
            raise e

    # For backward compatibility
    def abort(self):
        self.warning("abort is not implemented, stop is being called instead")
        self.stop()

    def stop(self):
        '''stop the macro.'''
        if self.door is None:
            return
        self.door.PauseMacro()
        # Since this could be done by error (impatient users clicking more than once)
        # we provide a warning message that does not make the process too slow
        # It may also be useful and 'ABORT' at TaurusApplication level
        # (macros+motions+acquisitions)
        title = 'Stopping macro'
        message = 'The following macro is still running:\n\n'
        message += '%s %s\n\n' % (self.macro_name, ' '.join(self.macro_args))
        message += 'Are you sure you want to stop?\n'
        buttons = Qt.QMessageBox.Ok | Qt.QMessageBox.Cancel
        ans = Qt.QMessageBox.warning(
            self, title, message, buttons, Qt.QMessageBox.Ok)
        if ans == Qt.QMessageBox.Ok:
            self.door.stop(synch=True)
        else:
            self.ui.button.setChecked(True)
            self.door.ResumeMacro()

    @classmethod
    def getQtDesignerPluginInfo(cls):
        '''reimplemented from :class:`TaurusWidget`'''
        return {'container': False,
                'group': 'Taurus Sardana',
                'module': 'taurus.qt.qtgui.extra_macroexecutor',
                'icon': ':/designer/pushbutton.png'}


class MacroButtonAbortDoor(Qt.QPushButton, TaurusBaseWidget):
    '''Deprecated class. Instead use TaurusCommandButton.
    A button for aborting macros on a door
    '''
    # todo: why not inheriting from (TaurusBaseComponent, Qt.QPushButton)?

    def __init__(self, parent=None, designMode=False):
        name = self.__class__.__name__
        self.call__init__wo_kw(Qt.QPushButton, parent)
        self.call__init__(TaurusBaseWidget, name, designMode=designMode)
        self.warning('Deprecation warning: use TaurusCommandButton class ' +
                     'instead of MacroButtonAbortDoor')

        self.setText('Abort')
        self.setToolTip('Abort Macro')
        self.clicked.connect(self.abort)

    def getModelClass(self):
        '''reimplemented from :class:`TaurusBaseWidget`'''
        return TaurusDevice

    @ProtectTaurusMessageBox(msg='An error occurred trying to abort the macro.')
    def abort(self):
        '''stops macros'''
        door = self.getModelObj()
        if door is not None:
            door.stopMacro()


if __name__ == '__main__':
    import sys
    from taurus.qt.qtgui.application import TaurusApplication
    from taurus.core.util.argparse import get_taurus_parser
    from sardana.taurus.qt.qtcore.tango.sardana.macroserver import \
        registerExtensions
    registerExtensions()

    parser = get_taurus_parser()
    parser.set_usage("python macrobutton.py [door_name] [macro_name]")
    parser.set_description("Macro button for macro execution")

    app = TaurusApplication(app_name="macrobutton",
                            app_version=taurus.Release.version,
                            cmd_line_parser=parser)

    args = app.get_command_line_args()

    nargs = len(args)

    if nargs == 1:
        macro_name = 'lsmac'
    elif nargs == 2:
        macro_name = args[1]
    else:
        parser.print_help(sys.stderr)
        sys.exit(1)
    door_name = args[0]

    class TestWidget(Qt.QWidget):

        def __init__(self, door, macro):
            Qt.QWidget.__init__(self)

            self.door_name = door
            self.macro_name = macro
            self.w_macro_name = None
            self.create_layout(macro)

        def update_macro_name(self, macro_name):
            self.macro_name = macro_name

        def update_result(self, result):
            self.result_label.setText(str(result))

        def toggle_progress(self, toggle):
            visible = self.show_progress.isChecked()
            self.mb.showProgress(visible or toggle)

        def getMacroInfo(self, macro_name):

            door = taurus.Device(self.door_name)
            try:
                pars = door.macro_server.getMacroInfoObj(macro_name).parameters
            except AttributeError as e:
                print("Macro %s does not exists!" % macro_name)
                return None

            param_names = []
            default_values = []
            for p in pars:
                ptype = p['type']

                if isinstance(ptype, list):
                    for pr in ptype:
                        param_names.append(pr['name'])
                        default_values.append(pr['default_value'])
                else:
                    param_names.append(p['name'])
                    default_values.append(p['default_value'])

            self.macro_name = macro_name
            return param_names, default_values

        def clean_layout(self, layout):

            while layout.count():
                child = layout.takeAt(0)
                if child.widget():
                    child.widget().deleteLater()

        def create_layout(self, macro_name):
            p_names, d_values = self.getMacroInfo(macro_name)
            # Create TOP layout
            self.w_arg = Qt.QWidget()
            self.w_arg.setLayout(Qt.QGridLayout())
            col = 0
            self.w_arg.layout().addWidget(Qt.QLabel('macro name'), 0, col)
            self.w_macro_name = Qt.QLineEdit()
            self.w_arg.layout().addWidget(self.w_macro_name, 1, col)

            _argEditors = []
            for name in p_names:
                col += 1
                self.w_arg.layout().addWidget(Qt.QLabel(name), 0, col)
                self.argEdit = Qt.QLineEdit()
                self.w_arg.layout().addWidget(self.argEdit, 1, col)
                _argEditors.append(self.argEdit)

            for e, v in zip(_argEditors, d_values):
                if v is None:
                    continue
                e.setText(str(v))

            # Create bottom layout
            self.mb = MacroButton()
            self.mb.setModel(door_name)
            self.w_bottom = Qt.QWidget()
            self.w_bottom.setLayout(Qt.QGridLayout())
            self.w_bottom.layout().addWidget(self.mb, 0, 0, 2, 7)
            self.w_bottom.layout().addWidget(Qt.QLabel('Result:'), 2, 0)

            self.result_label = Qt.QLabel()
            self.w_bottom.layout().addWidget(self.result_label, 2, 1, 1, 5)

            self.show_progress = Qt.QCheckBox('Progress')
            self.show_progress.setChecked(True)
            self.w_bottom.layout().addWidget(self.show_progress, 3, 0)

            mb_abort = TaurusCommandButton(command='StopMacro',
                                           icon=':/actions/media_playback_stop.svg')
            mb_abort.setModel(door_name)
            self.w_bottom.layout().addWidget(mb_abort, 3, 1)

            # Toggle progressbar
            self.show_progress.stateChanged.connect(self.toggle_progress)
            # connect the argument editors
            for i, editor in enumerate(_argEditors):
                slot = functools.partial(self.mb.updateMacroArgument, i)
                editor.textChanged.connect(slot)

            self.setLayout(Qt.QVBoxLayout())
            self.layout().addWidget(self.w_arg)
            self.layout().addWidget(self.w_bottom)

            # Update possible macro result
            self.mb.resultUpdated.connect(self.update_result)

            self.w_macro_name.textEdited.connect(self.update_macro_name)
            self.w_macro_name.editingFinished.connect(self.update_layout)
            self.w_macro_name.textChanged.connect(self.mb.setMacroName)
            self.w_macro_name.textChanged.connect(self.mb.setButtonText)

            # Since everything is now connected, the parameters will be updated
            self.w_macro_name.setText(macro_name)

        def update_layout(self):
            if self.getMacroInfo(self.macro_name):
                self.clean_layout(self.layout())
                self.create_layout(self.macro_name)

    w = TestWidget(door_name, macro_name)
    w.show()
    sys.exit(app.exec_())
